from urllib.parse import urlparse

from pyasn1_alt_modules import rfc5280

from pkilint import validation, common
from pkilint.pkix import extension


class AuthorityInformationAccessPresenceValidator(extension.ExtensionPresenceValidator):
    _VALIDATION_TYPE = "cabf.aia_extension_missing"

    def __init__(self, severity):
        finding = validation.ValidationFinding(severity, self._VALIDATION_TYPE)

        super().__init__(
            extension_oid=rfc5280.id_pe_authorityInfoAccess,
            validation=finding,
            pdu_class=rfc5280.Extensions,
        )


def _general_name_is_http_uri(node):
    name, value = node.child

    return name == "uniformResourceIdentifier" and str(value.pdu).lower().startswith(
        "http:"
    )


def _validate_descriptions(node, access_method, empty_finding, no_http_finding):
    ads = [
        ad
        for ad in node.children.values()
        if ad.children["accessMethod"].pdu == access_method
    ]

    if len(ads) == 0:
        if empty_finding is None:
            return
        else:
            raise validation.ValidationFindingEncountered(empty_finding)

    if not any(
        filter(lambda a: _general_name_is_http_uri(a.children["accessLocation"]), ads)
    ):
        raise validation.ValidationFindingEncountered(no_http_finding)


class AuthorityInformationAccessContainsHttpUriValidator(validation.Validator):
    VALIDATION_NO_CA_ISSUERS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING, "cabf.aia_ca_issuers_missing"
    )

    VALIDATION_CA_ISSUERS_NO_HTTP_URI = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.aia_ca_issuers_has_no_http_uri",
    )

    VALIDATION_OCSP_NO_HTTP_URI = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "cabf.aia_ocsp_has_no_http_uri"
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_NO_CA_ISSUERS,
                self.VALIDATION_CA_ISSUERS_NO_HTTP_URI,
                self.VALIDATION_OCSP_NO_HTTP_URI,
            ],
            pdu_class=rfc5280.AuthorityInfoAccessSyntax,
        )

    def validate(self, node):
        _validate_descriptions(
            node,
            rfc5280.id_ad_caIssuers,
            self.VALIDATION_NO_CA_ISSUERS,
            self.VALIDATION_CA_ISSUERS_NO_HTTP_URI,
        )
        _validate_descriptions(
            node, rfc5280.id_ad_ocsp, None, self.VALIDATION_OCSP_NO_HTTP_URI
        )


class CrlDpContainsHttpUriValidator(validation.Validator):
    VALIDATION_CRLDP_NO_HTTP_URI = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "cabf.no_http_crldp_uri"
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_CRLDP_NO_HTTP_URI],
            pdu_class=rfc5280.CRLDistributionPoints,
        )

    def validate(self, node):
        for dist_point in node.children.values():
            dp = dist_point.children.get("distributionPoint")

            if dp is None:
                continue

            full_name = dp.children.get("fullName")
            if full_name is None:
                continue

            for gn in full_name.children.values():
                name, value = gn.child
                if name == "uniformResourceIdentifier" and str(
                    value.pdu
                ).lower().startswith("http:"):
                    return

        raise validation.ValidationFindingEncountered(self.VALIDATION_CRLDP_NO_HTTP_URI)


class CertificatePoliciesCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_CERTIFICATE_POLICIES_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "cabf.critical_certificate_policies_extension",
    )

    def __init__(self):
        super().__init__(
            validation=self.VALIDATION_CERTIFICATE_POLICIES_CRITICAL,
            type_oid=rfc5280.id_ce_certificatePolicies,
            is_critical=False,
        )


class CabfCrlDpCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_CRLDP_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "cabf.critical_crldp_extension"
    )

    def __init__(self):
        super().__init__(
            validation=self.VALIDATION_CRLDP_CRITICAL,
            type_oid=rfc5280.id_ce_cRLDistributionPoints,
            is_critical=False,
        )


class CabfAuthorityKeyIdentifierValidator(validation.Validator):
    VALIDATION_AKI_HAS_ISSUER_CERT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.authority_key_identifier_has_issuer_cert",
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_AKI_HAS_ISSUER_CERT],
            pdu_class=rfc5280.AuthorityKeyIdentifier,
        )

    def validate(self, node):
        if (
            "authorityCertIssuer" in node.children
            or "authorityCertSerialNumber" in node.children
        ):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_AKI_HAS_ISSUER_CERT
            )


class CabfExtensionsPresenceValidator(common.ExtensionsPresenceValidator):
    VALIDATION_EXTENSIONS_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.certificate_extensions_missing",
    )

    def __init__(self):
        super().__init__(self.VALIDATION_EXTENSIONS_MISSING)


class CpsUriHttpValidator(validation.Validator):
    VALIDATION_CPS_URI_NOT_HTTP = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "cabf.cps_uri_is_not_http"
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_CPS_URI_NOT_HTTP], pdu_class=rfc5280.CPSuri
        )

    def validate(self, node):
        uri = str(node.pdu)
        scheme = urlparse(uri).scheme

        if scheme.lower() not in {"http", "https"}:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CPS_URI_NOT_HTTP, f'Prohibited URI scheme: "{scheme}"'
            )



import codecs
import ipaddress
import math
from urllib.parse import urlparse

import validators
from pyasn1_alt_modules import rfc5280, rfc8398

from pkilint import validation

_GENERALNAME_INSTANCE = rfc5280.GeneralName()
_GENERALNAME_TYPES = [str(n) for n in _GENERALNAME_INSTANCE.componentType]

OTHER_NAME_MAPPINGS = rfc5280.anotherNameMap.copy()


# TODO: consider subclassing from StrEnum when minimum supported Python version is 3.11
class GeneralNameTypeName:
    OTHER_NAME = "otherName"
    RFC822_NAME = "rfc822Name"
    DNS_NAME = "dNSName"
    X400_ADDRESS = "x400Address"
    DIRECTORY_NAME = "directoryName"
    EDI_PARTY_NAME = "ediPartyName"
    UNIFORM_RESOURCE_IDENTIFIER = "uniformResourceIdentifier"
    IP_ADDRESS = "iPAddress"
    REGISTERED_ID = "registeredID"


def validators_predicate(func, value):
    ret = func(value)

    return isinstance(ret, bool) and ret


def is_nameconstraints_child_node(node):
    return any((isinstance(n.pdu, rfc5280.NameConstraints) for n in node.parents))


def create_generalname_type_predicate(generalname_type):
    if generalname_type in _GENERALNAME_TYPES:
        return lambda n: (
            n.name == generalname_type
            and n.parent is not None
            and isinstance(n.parent.pdu, rfc5280.GeneralName)
        )
    else:
        raise ValueError(f"Invalid type specified: {generalname_type}")


class GeneralNameValidatorContainer(validation.ValidatorContainer):
    def __init__(self):
        super().__init__(
            validators=[
                GeneralNameUriSyntaxValidator(),
                GeneralNameDnsNameSyntaxValidator(),
                GeneralNameIpAddressSyntaxValidator(),
                GeneralNameMailboxAddressSyntaxValidator(),
                SmtpUTF8MailboxValidator(),
                GeneralNameDnsNameDomainNameLengthValidator(),
                GeneralNameUriDomainNameLengthValidator(),
                GeneralNameRfc822NameDomainNameLengthValidator(),
                SmtpUTF8MailboxDomainNameLengthValidator(),
            ],
            pdu_class=rfc5280.GeneralName,
        )


class UriSyntaxValidator(validation.Validator):
    VALIDATION_INVALID_URI_SYNTAX = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.invalid_uri_syntax"
    )

    VALIDATION_LDAP_URI_NOT_VALIDATED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE, "pkix.ldap_uri_not_validated"
    )

    def __init__(self, **kwargs):
        super().__init__(
            validations=[
                self.VALIDATION_INVALID_URI_SYNTAX,
                self.VALIDATION_LDAP_URI_NOT_VALIDATED,
            ],
            **kwargs,
        )

    def validate_value(self, node):
        value = str(node.pdu)

        if value.casefold().startswith("ldap://".casefold()):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_LDAP_URI_NOT_VALIDATED, value
            )

        return validators_predicate(validators.url, value)

    def validate(self, node):
        if not self.validate_value(node):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INVALID_URI_SYNTAX,
                f'Invalid URI syntax: "{str(node.pdu)}"',
            )


class DomainNameSyntaxValidator(validation.Validator):
    VALIDATION_NOT_PREFERRED_NAME_SYNTAX = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.invalid_domain_name_syntax"
    )

    def __init__(self, **kwargs):
        super().__init__(
            validations=[self.VALIDATION_NOT_PREFERRED_NAME_SYNTAX], **kwargs
        )

    def validate_value(self, node):
        value = str(node.pdu)

        return validators_predicate(validators.domain, value)

    def validate(self, node):
        if not self.validate_value(node):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_NOT_PREFERRED_NAME_SYNTAX,
                f'Invalid domain name syntax: "{str(node.pdu)}"',
            )


class MailboxAddressSyntaxValidator(validation.Validator):
    VALIDATION_INVALID_EMAIL_ADDRESS_SYNTAX = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.invalid_email_address_syntax"
    )

    def __init__(self, **kwargs):
        super().__init__(
            validations=[self.VALIDATION_INVALID_EMAIL_ADDRESS_SYNTAX], **kwargs
        )

    def validate_value(self, node):
        value = str(node.pdu)

        return validators_predicate(validators.email, value)

    def validate(self, node):
        if not self.validate_value(node):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INVALID_EMAIL_ADDRESS_SYNTAX,
                f'Invalid e-mail address syntax: "{str(node.pdu)}"',
            )


class GeneralNameUriSyntaxValidator(UriSyntaxValidator):
    def __init__(self):
        super().__init__(
            predicate=create_generalname_type_predicate("uniformResourceIdentifier"),
        )

    def validate_value(self, node):
        is_nc_child = is_nameconstraints_child_node(node)

        value = str(node.pdu)

        if is_nc_child:
            if value == ".":
                return True

            if value.startswith("."):
                value = value[1:]

            return validators_predicate(validators.domain, value)
        else:
            return super().validate_value(node)


class GeneralNameDnsNameSyntaxValidator(DomainNameSyntaxValidator):
    def __init__(self):
        super().__init__(predicate=create_generalname_type_predicate("dNSName"))

    def validate_value(self, node):
        value = str(node.pdu)

        if len(value) == 0 and is_nameconstraints_child_node(node):
            return True
        else:
            return super().validate_value(node)


class GeneralNameMailboxAddressSyntaxValidator(MailboxAddressSyntaxValidator):
    def __init__(self):
        super().__init__(
            predicate=create_generalname_type_predicate("rfc822Name"),
        )

    def validate_value(self, node):
        value = str(node.pdu)

        if "@" in value:
            return super().validate_value(node)

        if is_nameconstraints_child_node(node):
            if value == ".":
                return True

            if value.startswith("."):
                value = value[1:]

            return validators_predicate(validators.domain, value)
        else:
            return super().validate_value(node)


def _get_cidr_prefix(address_octets):
    address_bitlen = len(address_octets) * 8

    address = int.from_bytes(address_octets, "big")

    if address == 0:
        return 0

    lsb = int(math.log2(address & -address))

    for i in range(address_bitlen - 1, lsb, -1):
        if address & (1 << i) == 0:
            raise ValueError(f"Netmask bit index {i} is not asserted")

    return address_bitlen - lsb


class GeneralNameIpAddressSyntaxValidator(validation.Validator):
    VALIDATION_IP_ADDRESS_NC_WRONG_LENGTH = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.ip_address_name_constraint_wrong_length",
    )

    VALIDATION_IP_ADDRESS_WRONG_LENGTH = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.ip_address_wrong_length"
    )

    VALIDATION_IP_ADDRESS_CONSTRAINT_NOT_CIDR = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.ip_address_name_constraint_invalid_cidr",
    )

    def __init__(self):
        super().__init__(
            predicate=create_generalname_type_predicate("iPAddress"),
            validations=[
                self.VALIDATION_IP_ADDRESS_NC_WRONG_LENGTH,
                self.VALIDATION_IP_ADDRESS_WRONG_LENGTH,
                self.VALIDATION_IP_ADDRESS_CONSTRAINT_NOT_CIDR,
            ],
        )

    def validate(self, node):
        is_nc_child = is_nameconstraints_child_node(node)

        length_multiplier = 2 if is_nc_child else 1

        allowed_lengths = [
            address_lengths * length_multiplier for address_lengths in [4, 16]
        ]

        value = node.pdu.asOctets()

        actual_length = len(value)

        if actual_length not in allowed_lengths:
            if is_nc_child:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_IP_ADDRESS_NC_WRONG_LENGTH,
                    f"Invalid IP address contraint length: {actual_length}",
                )
            else:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_IP_ADDRESS_WRONG_LENGTH,
                    f"Invalid IP address length: {actual_length}",
                )

        if is_nc_child:
            midpoint = int(actual_length / 2)

            network = value[0:midpoint]
            mask = value[midpoint:]

            try:
                cidr = _get_cidr_prefix(mask)

                if actual_length == (4 * 2):
                    network_obj = ipaddress.IPv4Address(network)

                    ipaddress.IPv4Network(f"{network_obj}/{cidr}")
                else:
                    network_obj = ipaddress.IPv6Address(network)

                    ipaddress.IPv6Network(f"{network_obj}/{cidr}")
            except ValueError as e:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_IP_ADDRESS_CONSTRAINT_NOT_CIDR, str(e)
                )


class SmtpUTF8MailboxValidator(validation.Validator):
    VALIDATION_NOT_MAILBOX_ADDRESS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.smtp_utf8_mailbox_invalid_syntax",
    )

    VALIDATION_ASCII_ONLY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.smtp_utf8_mailbox_is_ascii_only",
    )

    VALIDATION_HAS_BOM = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.smtp_utf8_mailbox_has_bom"
    )

    VALIDATION_DOMAIN_PART_NOT_LOWERCASE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.smtp_utf8_mailbox_has_uppercase",
    )

    """
    RFC 9598, section 3:
     In SmtpUTF8Mailbox, labels that include non-ASCII characters MUST be stored in A-label (rather than U-label)
     form [RFC5890].
    """
    VALIDATION_DOMAIN_PART_INVALID_DOMAIN_SYNTAX = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.smtp_utf8_mailbox_domain_part_invalid_syntax",
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_NOT_MAILBOX_ADDRESS,
                self.VALIDATION_ASCII_ONLY,
                self.VALIDATION_HAS_BOM,
                self.VALIDATION_DOMAIN_PART_NOT_LOWERCASE,
                self.VALIDATION_DOMAIN_PART_INVALID_DOMAIN_SYNTAX,
            ],
            pdu_class=rfc8398.SmtpUTF8Mailbox,
        )

    def validate(self, node):
        value = str(node.pdu)

        if "@" not in value:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_NOT_MAILBOX_ADDRESS
            )

        local_part, domain_part = value.split("@", maxsplit=1)

        if local_part.isascii():
            raise validation.ValidationFindingEncountered(self.VALIDATION_ASCII_ONLY)

        encoded = local_part.encode("utf-8")
        if encoded.startswith(codecs.BOM_UTF8):
            raise validation.ValidationFindingEncountered(self.VALIDATION_HAS_BOM)

        if domain_part != domain_part.lower():
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_DOMAIN_PART_NOT_LOWERCASE
            )

        if not domain_part.isascii() or not validators_predicate(
            validators.domain, domain_part
        ):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_DOMAIN_PART_INVALID_DOMAIN_SYNTAX
            )


class DomainNameLengthValidator(validation.Validator):
    # https://devblogs.microsoft.com/oldnewthing/20120412-00/?p=7873
    _MAX_DOMAIN_NAME_ASCII_LENGTH = 253

    def __init__(self, validation_domain_name_too_long, **kwargs):
        super().__init__(validations=[validation_domain_name_too_long], **kwargs)

        self._validation_domain_name_too_long = validation_domain_name_too_long

    @classmethod
    def extract_value(cls, node):
        pass

    def validate(self, node):
        value = self.extract_value(node)

        value_len = len(value)

        if value_len > self._MAX_DOMAIN_NAME_ASCII_LENGTH:
            raise validation.ValidationFindingEncountered(
                self._validation_domain_name_too_long,
                f'Domain name too long: "{value}" ({value_len} characters) exceeds maximum length of '
                f"{self._MAX_DOMAIN_NAME_ASCII_LENGTH} characters",
            )


class GeneralNameDnsNameDomainNameLengthValidator(DomainNameLengthValidator):
    VALIDATION_DOMAIN_NAME_TOO_LONG = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.dnsname_domain_name_too_long"
    )

    def __init__(self):
        super().__init__(
            self.VALIDATION_DOMAIN_NAME_TOO_LONG,
            predicate=create_generalname_type_predicate("dNSName"),
        )

    @classmethod
    def extract_value(cls, node):
        return str(node.pdu)


class GeneralNameUriDomainNameLengthValidator(DomainNameLengthValidator):
    VALIDATION_DOMAIN_NAME_TOO_LONG = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.uri_domain_name_too_long"
    )

    def __init__(self):
        super().__init__(
            self.VALIDATION_DOMAIN_NAME_TOO_LONG,
            predicate=create_generalname_type_predicate("uniformResourceIdentifier"),
        )

    @classmethod
    def extract_value(cls, node):
        uri_str = str(node.pdu)

        return urlparse(uri_str).hostname or ""


class GeneralNameEmailAddressDomainNameLengthValidator(DomainNameLengthValidator):
    def __init__(self, validation_domain_name_too_long, **kwargs):
        super().__init__(validation_domain_name_too_long, **kwargs)

    @classmethod
    def extract_value(cls, node):
        email_address = str(node.pdu)

        parts = email_address.split("@", maxsplit=1)

        return parts[1] if len(parts) == 2 else ""


class GeneralNameRfc822NameDomainNameLengthValidator(
    GeneralNameEmailAddressDomainNameLengthValidator
):
    VALIDATION_DOMAIN_NAME_TOO_LONG = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.rfc822name_domain_part_too_long",
    )

    def __init__(self):
        super().__init__(
            self.VALIDATION_DOMAIN_NAME_TOO_LONG,
            predicate=create_generalname_type_predicate("rfc822Name"),
        )


class SmtpUTF8MailboxDomainNameLengthValidator(
    GeneralNameEmailAddressDomainNameLengthValidator
):
    VALIDATION_DOMAIN_NAME_TOO_LONG = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.smtputf8mailbox_domain_part_too_long",
    )

    def __init__(self):
        super().__init__(
            self.VALIDATION_DOMAIN_NAME_TOO_LONG, pdu_class=rfc8398.SmtpUTF8Mailbox
        )



from pyasn1_alt_modules import rfc5280

from pkilint import validation
from pkilint.common import common_name
from pkilint.pkix import general_name, extension

_ALLOWED_GENERAL_NAME_TYPES = {general_name.GeneralNameTypeName.DNS_NAME}


class NcpWCommonNameValidator(common_name.CommonNameValidator):
    """
    WEB-4.1.3-4 (d):

    If necessary to distinguish the website identified by the subject name, the subject commonName may contain a
    domain name or a Wildcard Domain Name (as defined in BRG [9]) which is one of the dNSName values of
    the subjectAltName extension of a website authentication certificate.
    """

    VALIDATION_COMMON_NAME_UNKNOWN_SOURCE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_4.web-4.1.3-4.common_name_unknown_source",
    )

    def __init__(self):
        super().__init__(
            _ALLOWED_GENERAL_NAME_TYPES, self.VALIDATION_COMMON_NAME_UNKNOWN_SOURCE
        )


class QncpWCommonNameValidator(common_name.CommonNameValidator):
    """
    WEB-4.1.4-2:

    If necessary to distinguish the website identified by the subject name, the subject commonName may contain a
    domain name or a Wildcard Domain Name (as defined in BRG [9]) which is one of the dNSName values of
    the subjectAltName extension of a website authentication certificate.
    """

    VALIDATION_COMMON_NAME_UNKNOWN_SOURCE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_4.web-4.1.4-2.common_name_unknown_source",
    )

    def __init__(self):
        super().__init__(
            _ALLOWED_GENERAL_NAME_TYPES, self.VALIDATION_COMMON_NAME_UNKNOWN_SOURCE
        )


class NcpWExtendedKeyUsagePresenceValidator(extension.ExtensionPresenceValidator):
    """
    WEB-4.1.3-4: The following certificate profile requirements specified in the BRG [9] shall apply for subject
    certificate fields addressed by the following sub-sections of BRG [9] (the version of BRG [9] shall be as referenced
    in ETSI EN 319 411-1 [6] for [WEB] requirements):
    a) 7.1.2.3 f) extKeyUsage
    """

    VALIDATION_EKU_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_4.web-4.1.3-4.eku_missing",
    )

    def __init__(self):
        super().__init__(
            extension_oid=rfc5280.id_ce_extKeyUsage,
            validation=self.VALIDATION_EKU_MISSING,
            pdu_class=rfc5280.Extensions,
        )


class NcpWSubjectAltNamePresenceValidator(extension.ExtensionPresenceValidator):
    """
    WEB-4.1.3-4: The following certificate profile requirements specified in the BRG [9] shall apply for subject
    certificate fields addressed by the following sub-sections of BRG [9] (the version of BRG [9] shall be as referenced
    in ETSI EN 319 411-1 [6] for [WEB] requirements):
    a) 7.1.2.3 b) Subject Alternative Name
    """

    VALIDATION_SAN_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_4.web-4.1.3-4.san_missing",
    )

    def __init__(self):
        super().__init__(
            extension_oid=rfc5280.id_ce_subjectAltName,
            validation=self.VALIDATION_SAN_MISSING,
            pdu_class=rfc5280.Extensions,
        )


class NcpWCriticalityExtendedKeyUsageValidator(extension.ExtensionCriticalityValidator):
    """Validates that the criticality of the EKU extension conforms to BRG."""

    EXTENDED_KEY_USAGE_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_4.web-4.1.3-4.eku_extension_is_critical",
    )

    def __init__(self):
        super().__init__(
            validation=self.EXTENDED_KEY_USAGE_CRITICAL,
            type_oid=rfc5280.id_ce_extKeyUsage,
            is_critical=False,
        )



#!/usr/bin/env python

import argparse
import sys

from pkilint import loader, pkix, report, cli_util
from pkilint.cabf import cabf_crl
from pkilint.pkix import crl, name, extension


def _add_args(parser):
    parser.add_argument(
        "-t",
        "--type",
        required=True,
        type=str.upper,
        choices=["CRL", "ARL"],
        help="The type of CRL (whether the CRL in question is a CRL or an ARL)",
    )
    parser.add_argument(
        "-p",
        "--profile",
        required=True,
        type=str.upper,
        choices=["PKIX", "BR"],
        default="PKIX",
        help="The profile against which to lint",
    )


def main(cli_args=None) -> int:
    parser = argparse.ArgumentParser(description="RFC 5280 and CA/B Forum CRL Linter")

    subparsers = parser.add_subparsers(dest="command", required=True)
    validations_parser = subparsers.add_parser(
        "validations", help="Output the set of validations which this linter performs"
    )
    _add_args(validations_parser)

    lint_parser = subparsers.add_parser("lint", help="Lint the specified CRL")
    _add_args(lint_parser)
    cli_util.add_standard_args(lint_parser)

    lint_parser.add_argument(
        "file", type=argparse.FileType("rb"), help="The CRL file to lint"
    )

    args = parser.parse_args(cli_args)

    crl_type = crl.CertificateRevocationListType[args.type]

    validity_additional_validators = []
    doc_additional_validators = []

    if args.profile == "BR":
        doc_additional_validators.append(
            cabf_crl.CabfCrlReasonCodeAllowlistValidator(crl_type)
        )

        validity_additional_validators.append(
            cabf_crl.create_validity_period_validator(crl_type)
        )

    doc_validator = crl.create_pkix_crl_validator_container(
        [
            pkix.create_attribute_decoder(name.ATTRIBUTE_TYPE_MAPPINGS),
            pkix.create_extension_decoder(extension.EXTENSION_MAPPINGS),
        ],
        [
            crl.create_issuer_validator_container([]),
            crl.create_validity_validator_container(validity_additional_validators),
            crl.create_extensions_validator_container([]),
        ]
        + doc_additional_validators,
    )

    if args.command == "validations":
        print(report.report_included_validations(doc_validator))

        return 0
    else:
        try:
            crl_doc = (
                loader.RFC5280CertificateListDocumentLoader().get_file_loader_func(
                    args.document_format
                )(args.file, args.file.name)
            )
        except ValueError as e:
            print(f"Failed to load CRL: {e}", file=sys.stderr)
            return 1

        results = doc_validator.validate(crl_doc.root)

        print(args.format(results, args.severity))

        return cli_util.clamp_exit_code(
            report.get_findings_count(results, args.severity)
        )


if __name__ == "__main__":
    sys.exit(main())



from pyasn1_alt_modules import rfc5280

from pkilint import validation, document
from pkilint.pkix import certificate, general_name


class EndEntityRevocationInformationPresenceValidator(validation.Validator):
    """
    Microsoft Root Program Requirements, 3.A.5:

    An end-entity certificate may contain either an AIA extension with a valid OCSP URL and/or a CDP extension pointing
    to a valid HTTP endpoint containing the CRL.
    """

    VALIDATION_REVOCATION_INFORMATION_ABSENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "msft.end_entity.revocation_information_absent",
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_REVOCATION_INFORMATION_ABSENT],
            pdu_class=rfc5280.Extensions,
        )

    @classmethod
    def _general_name_is_http_uri(cls, gn):
        gn_type, gn_value = gn.child

        return (
            gn_type == general_name.GeneralNameTypeName.UNIFORM_RESOURCE_IDENTIFIER
            and str(gn_value.pdu).lower().startswith("http://")
        )

    @classmethod
    def _has_ocsp_http_uri(cls, cert: certificate.RFC5280Certificate):
        aia_ext_and_idx = cert.get_extension_by_oid(rfc5280.id_pe_authorityInfoAccess)

        if aia_ext_and_idx is None:
            return False

        aia_ext, _ = aia_ext_and_idx

        # ensure that the decoded value is present
        try:
            aia_ext_value = aia_ext.navigate("extnValue.authorityInfoAccessSyntax")
        except document.PDUNavigationFailedError:
            return False

        ocsp_gns = (
            ad.children["accessLocation"]
            for ad in aia_ext_value.children.values()
            if ad.children["accessMethod"].pdu == rfc5280.id_ad_ocsp
        )

        return any(cls._general_name_is_http_uri(gn) for gn in ocsp_gns)

    @classmethod
    def _has_crldp_http_uri(cls, cert: certificate.RFC5280Certificate):
        crldp_ext_and_idx = cert.get_extension_by_oid(
            rfc5280.id_ce_cRLDistributionPoints
        )

        if crldp_ext_and_idx is None:
            return False

        crldp_ext, _ = crldp_ext_and_idx

        # ensure that the decoded value is present
        try:
            crldp_ext_value = crldp_ext.navigate("extnValue.cRLDistributionPoints")
        except document.PDUNavigationFailedError:
            return False

        for dp in crldp_ext_value.children.values():
            dpn = dp.children.get("distributionPoint")

            if dpn is None:
                continue

            full_name = dpn.children.get("fullName")

            if full_name is None:
                continue

            if any(
                cls._general_name_is_http_uri(gn) for gn in full_name.children.values()
            ):
                return True

        return False

    def validate(self, node):
        cert_doc = node.document

        has_aia_ocsp_http_uri = self._has_ocsp_http_uri(cert_doc)
        has_crldp_http_uri = self._has_crldp_http_uri(cert_doc)

        if not has_aia_ocsp_http_uri and not has_crldp_http_uri:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_REVOCATION_INFORMATION_ABSENT
            )



import binascii

from pyasn1.codec.der.encoder import encode
from pyasn1.type import univ
from pyasn1_alt_modules import rfc4055, rfc5280, rfc5480, rfc8410

from pkilint import validation, document
from pkilint.nist.asn1 import fips_204, fips_205

SIGNATURE_ALGORITHM_IDENTIFIER_MAPPINGS = {
    **{
        o: document.ValueDecoder.VALUE_NODE_ABSENT
        for o in (
            rfc8410.id_Ed448,
            rfc8410.id_Ed25519,
            rfc8410.id_X448,
            rfc8410.id_X25519,
            rfc5480.id_dsa_with_sha1,
            rfc5480.id_dsa_with_sha224,
            rfc5480.id_dsa_with_sha256,
            rfc5480.ecdsa_with_SHA1,
            rfc5480.ecdsa_with_SHA224,
            rfc5480.ecdsa_with_SHA256,
            rfc5480.ecdsa_with_SHA384,
            rfc5480.ecdsa_with_SHA512,
        )
    },
    **{
        o: univ.Null()
        for o in (
            rfc5480.md2WithRSAEncryption,
            rfc5480.md5WithRSAEncryption,
            rfc5480.sha1WithRSAEncryption,
            rfc4055.sha224WithRSAEncryption,
            rfc4055.sha256WithRSAEncryption,
            rfc4055.sha384WithRSAEncryption,
            rfc4055.sha512WithRSAEncryption,
        )
    },
    **fips_204.ALGORITHM_OID_TO_PARAMETER_MAPPINGS,
    **fips_205.ALGORITHM_OID_TO_PARAMETER_MAPPINGS,
    rfc4055.id_RSASSA_PSS: rfc4055.RSASSA_PSS_params(),
}


class AlgorithmIdentifierDecodingValidator(validation.DecodingValidator):
    def __init__(self, *, decode_func, **kwargs):
        super().__init__(
            pdu_class=rfc5280.AlgorithmIdentifier, decode_func=decode_func, **kwargs
        )


class AllowedSignatureAlgorithmEncodingValidator(validation.Validator):
    def __init__(self, *, validation, allowed_encodings, **kwargs):
        self._allowed_encodings = allowed_encodings

        super().__init__(validations=[validation], **kwargs)

    def validate(self, node):
        encoded = encode(node.pdu)

        if encoded not in self._allowed_encodings:
            encoded_str = binascii.hexlify(encoded).decode("us-ascii")

            raise validation.ValidationFindingEncountered(
                self._validations[0], f"Prohibited encoding: {encoded_str}"
            )
